/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2010
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_COMMON_COORDINATES_HPP__
#define YAYI_COMMON_COORDINATES_HPP__

#include <Yayi/core/yayiCommon/common_types.hpp>
#include <Yayi/core/yayiCommon/common_errors.hpp>
#include <vector>


namespace yayi
{
  // Forward declarations
  template <int dim>  struct s_coordinate;
  template <>         struct s_coordinate<0>;

  /*!@rief Main coordinates type
   *
   * @author Raffi Enficiaud
   */
  template <int dim> struct s_coordinate
  {
  public:

    typedef scalar_coordinate scalar_coordinate_type;		//!< Type contained in each dimension
    enum {static_dimensions = dim};

  private:
    scalar_coordinate coords[dim];
    typedef s_coordinate<dim> this_type;

  public:
    s_coordinate()
    {
      //for(int i = 0; i < dim; i++) coords[i] = r.coords[i];
    }
    s_coordinate(const this_type& r)
    {
      for(int i = 0; i < dim; i++) coords[i] = r.coords[i];
    }
    explicit s_coordinate(const scalar_coordinate_type r)
    {
      for(int i = 0; i < dim; i++) coords[i] = r;
    }
    /*explicit s_coordinate(const scalar_coordinate_type r[dim])
    {
      for(int i = 0; i < dim; i++) coords[i] = r[i];
    }*/

    explicit s_coordinate(const s_coordinate<0>& r);
#if 0
    {
      int i = 0;
      for(int j = std::min(r.dimension(), dim); i < j; i++) coords[i] = r[i];
      if(i < dim)
      {
        for(; i < dim; i++)
        {
          coords[i] = 0;
        }
      }
      else
      {
        for(int j = r.dimension(); i < j; i++)
        {
          DEBUG_ASSERT(
            r[i] == 0, 
            "Trying to copy a generic coordinate of dimension " + int_to_string(r.dimension())
            + std::string(" into a coordinate of dimension ") + int_to_string(dim));
        }
      }
    }
#endif

    static this_type from_table(const scalar_coordinate_type r[dim])
    {
      this_type out;
      for(int i = 0; i < dim; i++) out.coords[i] = r[i];
      return out;
    }

    static std::vector<this_type> from_table_multiple(const scalar_coordinate_type *r_begin, const scalar_coordinate_type *r_end)
    {
      DEBUG_ASSERT(((r_end - r_begin) % dim) == 0, "Incorrect number of elements");
      std::vector<this_type> out;
      for(; r_begin != r_end; r_begin += dim) out.push_back(from_table(r_begin));
      return out;
    }


    this_type& operator=(const this_type& r)
    {
      for(int i = 0; i < dim; i++) coords[i] = r.coords[i];
      return *this;
    }

    //! Comparison between two instances of coordinates
    bool operator==(const this_type& r) const throw()
    {
      for(int i = 0; i < dim; i++) if(coords[i] != r.coords[i]) return false;
      return true;
    }
    //! Comparison between two instances of coordinates
    bool operator!=(const this_type& r) const throw()
    {
      for(int i = 0; i < dim; i++) if(coords[i] != r.coords[i]) return true;
      return false;
    }
    
    /*! Simple lexicographical order between two coordinates of the same type
     *  Intended for convenience
     */
    bool operator<(const this_type& r) const throw() {
      for(int i = 0; i < dim; i++)
      {
        if(coords[i] == r.coords[i])
          continue;
        return coords[i] < r.coords[i];
      }
      return false;
    }

    //! Compare the coordinates to a scalar value
    bool operator==(const scalar_coordinate_type v) const throw()
    {
      for(int i = 0; i < dim; i++) if(coords[i] != v) return false;
      return true;
    }

    //! Compare the coordinates to a scalar value
    this_type operator+(const this_type& r_) const throw()
    {
      this_type return_(*this);
      for(int i = 0; i < dim; i++) return_.coords[i] += r_.coords[i];
      return return_;
    }


    //! Returns true if the coordinate is set to the origin of the space
    bool is_origin() const throw()
    {
      for(int i = 0; i < dim; i++) if(coords[i] != 0) return false;
      return true;
    }


    scalar_coordinate_type operator[](unsigned int i) const
    {
      DEBUG_ASSERT(i < dim, "Error out of range");
      return coords[i];
    }
    scalar_coordinate_type& operator[](unsigned int i)
    {
      DEBUG_ASSERT(i < dim, "Error out of range");
      return coords[i];
    }

    void clear() throw()
    {
      for(int i = 0; i < dim; i++) coords[i] = 0;
    }


    //! The coordinate dimension
    const unsigned int dimension() const throw() {return static_cast<unsigned int>(dim);}
  };


  /*!@brief Specializing of coordinates for unknown (any) dimension
  *
  * @author Raffi Enficiaud
  */
  template <> struct s_coordinate<0>
  {

    //! Storage class of the coordinates
    typedef scalar_coordinate scalar_coordinate_type;

  private:
    std::vector<scalar_coordinate_type> coords;
    typedef s_coordinate<0> this_type;

  public:

    //! Default constructor
    s_coordinate() : coords(){}


    //! Copy constructor
    s_coordinate(const this_type& r) : coords()
    {
      coords.assign(r.coords.begin(), r.coords.end());
    }

    //! Copy constructor from a coordinate of diffferent dimension
    template <int i_dim>
      s_coordinate(const s_coordinate<i_dim>& r) : coords()
    {
      coords.resize(r.dimension());
      for(int i = 0, j = r.dimension(); i < j; i++)
        coords[i] = r[i];
    }



    //! Comparison between two instances of coordinates
    bool operator==(const this_type& r) const throw()
    {
      if(r.dimension() != dimension())
        return false;
      for(int i = 0, j = dimension(); i < j; i++) if(coords[i] != r.coords[i]) return false;
      return true;
    }
    //! Comparison between two instances of coordinates
    bool operator!=(const this_type& r) const throw()
    {
      if(r.dimension() != dimension())
        return true;
      for(int i = 0, j = dimension(); i < j; i++) if(coords[i] != r.coords[i]) return true;
      return false;
    }


    //! Compare the coordinates to a scalar value
    bool operator==(const scalar_coordinate_type v) const throw()
    {
      for(int i = 0, j = dimension(); i < j; i++) if(coords[i] != v) return false;
      return true;
    }

    //! Returns true if the coordinate is set to the origin of the space
    bool is_origin() const throw()
    {
      for(int i = 0, j = dimension(); i < j; i++) if(coords[i] != 0) return false;
      return true;
    }



    //! Index operator (const)
    scalar_coordinate_type operator[](unsigned int i) const 
    {
      return coords[i];
    }

    //! Index operator (non const)
    scalar_coordinate_type& operator[](unsigned int i)
    {
      return coords[i];
    }

    //! Returns the number of dimensions (dynamic)
    const unsigned int dimension() const throw() {return static_cast<unsigned int>(coords.size());}

    //! Returns the number of dimensions (dynamic)
    void set_dimension(int d) throw() {coords.resize(d);}

    //! Resets the coordinate
    void clear() throw() {coords.clear();}

  };





  /*!@brief Rectangle type for a given dimension
  *
  */
  template <int dim> struct s_rectangle
  {
    typedef s_coordinate<dim> coordinate_type;

    coordinate_type upperleft_corner, lowerright_corner;
    
    //! Returns true if the specified point is inside the rectangle
    bool is_inside(const coordinate_type& c) const
    {
      for(int i = 0; i < dim; i++)
      {
        const typename coordinate_type::scalar_coordinate_type p = c[i];
        if(p < upperleft_corner[i] || p >= lowerright_corner[i])
          return false;
      }
      return true;
    }


  };
  

  template <int dim>
  s_coordinate<dim>::s_coordinate(const s_coordinate<0>& r)
  {
    unsigned int i = 0;
    for(unsigned int j = std::min(r.dimension(), static_cast<unsigned int>(dim)); i < j; i++) coords[i] = r[i];
    if(i < dim)
    {
      for(; i < dim; i++)
      {
        coords[i] = 0;
      }
    }
    else
    {
      for(unsigned int j = r.dimension(); i < j; i++)
      {
        DEBUG_ASSERT(
          r[i] == 0, 
          "Trying to copy a generic coordinate of dimension " + int_to_string(r.dimension())
          + std::string(" into a coordinate of dimension ") + int_to_string(dim));
      }
    }
  }

  inline s_coordinate<4> c4D(
    const s_coordinate<4>::scalar_coordinate_type a, 
    const s_coordinate<4>::scalar_coordinate_type b, 
    const s_coordinate<4>::scalar_coordinate_type c,
    const s_coordinate<4>::scalar_coordinate_type d) {
    
    const s_coordinate<4>::scalar_coordinate_type t[4] = {a,b,c,d};
    return s_coordinate<4>::from_table(t);
  } 

  inline s_coordinate<3> c3D(
    const s_coordinate<3>::scalar_coordinate_type x, 
    const s_coordinate<3>::scalar_coordinate_type y, 
    const s_coordinate<3>::scalar_coordinate_type z) {
    
    const s_coordinate<3>::scalar_coordinate_type t[3] = {x,y,z};
    return s_coordinate<3>::from_table(t);
  } 

  inline s_coordinate<2> c2D(
    const s_coordinate<2>::scalar_coordinate_type x, 
    const s_coordinate<2>::scalar_coordinate_type y) {
    
    const s_coordinate<2>::scalar_coordinate_type t[2] = {x,y};
    return s_coordinate<2>::from_table(t);
  } 

}


namespace std
{
  template <int T> inline std::ostream & operator<<(std::ostream &o, const yayi::s_coordinate<T> &c)
  {
    int i = 0, j = c.dimension() - 1;
    o << "(";
    for(; i < j; i++)
    {
      o << c[i] << ", ";
    }
    if(j > 0) o << c[j];
    o << ")";
    return o;
  }

  template <int T> inline std::wostream& operator<<(std::wostream &o, const yayi::s_coordinate<T> &c)
  {
    int i = 0, j = c.dimension() - 1;
    o << L"(";
    for(; i < j; i++)
    {
      o << c[i] << L", ";
    }
    if(j > 0) o << c[j];
    o << L")";
    return o;
  }

}

#endif


