/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2010
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_COMMON_TYPES_T_HPP__
#define YAYI_COMMON_TYPES_T_HPP__



#include <boost/mpl/bool.hpp>
#include <boost/type_traits.hpp>
#include <boost/mpl/comparison.hpp>
#include <boost/mpl/sizeof.hpp>
#include <boost/mpl/eval_if.hpp>
#include <boost/mpl/logical.hpp>
#include <boost/mpl/identity.hpp>


using namespace boost;


#include <Yayi/core/yayiCommon/common_types.hpp>
#include <Yayi/core/yayiCommon/common_errors.hpp>



namespace yayi
{
  
  namespace type_description
  {

    /*!@group meta_utility
     * @brief Metaprogramming utilities
     *
     *@{
     */


    /*!@brief Meta-utility class : always throws
     * Utility to prevent the instance creation of the daughter classes.
     *
     * @ingroup meta_utility
     * @author Raffi Enficiaud
     */
    struct s_always_throw
    {
      s_always_throw()
      {
        DEBUG_THROW("s_always_throw invoqued");
      }
    };

    /*!@brief Meta-utility class : never throws
     * @ingroup meta_utility
     *
     * This class does nothing particular (to oppose to s_always_throw)
     *
     * @author Raffi Enficiaud
     */
    struct s_never_throw{};


    /*!@brief Meta-utility class : type information
     * @ingroup meta_utility
     *
     * The purpose of this class is to prevent the use of unsupported types in the library. Each new supported type should be enquired by a
     * specializing of this class. 
     * 
     * @note Instances for unsupported types throw
     * @author Raffi Enficiaud
     */
    template <class U> struct type_support : public boost::mpl::false_
    {
      static const yayi::type::scalar_type scalar     = yayi::type::s_undefined;      //!< The scalar type
      static const yayi::type::compound_type compound = yayi::type::c_unknown;        //!< The compound type
      static const string_type& name() {
        static const string_type s = "";
        return s;
      }

      //typedef typename type_support<U>::generate_error type_error;
    };


    //! Utility transforming a template type to a type object
    //! @ref type
    template <class T> type to_type() {
      return type(type_support<T>::compound, type_support<T>::scalar);
    }




		/*!@brief Type description structure
		 *
     * The class U should be specialized by a specific @ref type_support
     * @author Raffi Enficiaud
		 */
		template <class U> 
      struct type_desc : 
        public mpl::eval_if<type_support<U>, mpl::identity<s_never_throw>, mpl::identity<s_always_throw> >::type,
        public type_support<U>
		{
#ifdef YAYI_COMPILER_ERROR_FOR_UNSUPPORTED_TYPES__
    private:
      //! The following typedef generates a compiler error for unsupported types instanciation
      typedef typename type_support<U>::support_value_type    generate_error;
#endif
    
    public:
      typedef boost::is_pod<U>                                is_pod;
      typedef typename boost::remove_cv<U>::type              type_remove_cv;
      typedef U                                               type;
		};

      





		//! Returns the most wide scalar type
		template <typename U, typename V>
			struct s_super_scalar
		{
			typedef typename boost::mpl::eval_if<
				typename boost::mpl::greater<
					boost::mpl::sizeof_<U>,
					boost::mpl::sizeof_<V> >::type,
				U,
				V>::type type;
		};



#if 1

		//! Returns the types that may encode both types
		template <class U, class V> struct supertype : 
			public mpl::if_<
				typename boost::mpl::and_<
					typename boost::is_scalar<U>::type, 
					typename boost::is_scalar<V>::type >::type,
				boost::mpl::identity<s_always_throw>,
				boost::mpl::identity<s_never_throw> 
				>::type
		{
			typedef typename s_super_scalar<U,V>::type type;
		};
#endif


		//! Returns an appropriate representation for summing a containt of type U
		template <typename U> struct s_sum_supertype {
			typedef U type;
		};



#ifdef SAFETY_FOR_ARITHMETIC
		template <typename U>
			typename s_sum_supertype<U>::type operator+(const U& l_, const U& r_)
		{
			U ret = l_;
			DEBUG_ASSERT(
				l_ + r_ > std::numeric_limits<>::max(), 
				"operator+ : out of bound for type " + std::string(type_desc<U>::name);
			ret += r_;
			return ret;
		}
#endif


	} // namespace type_description
} // namespace yayi

#endif /* YAYI_COMMON_TYPES_T_HPP__ */


