/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2010
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_APPLY_TO_IMAGE_UNARY_HPP__
#define YAYI_APPLY_TO_IMAGE_UNARY_HPP__

/*!@file
 * This file defines the unary operators on images.
 * @todo: add multithreading
 */


#include <Yayi/core/yayiImageCore/include/ApplyToImage_T.hpp>


namespace yayi
{

  //! Specializing of s_apply_op_range for independant iterators and unary operators
  template <class it_strategy /*= iterators_independant_tag*/>
  struct s_apply_op_range<it_strategy, operator_type_unary>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      for(; it1 != it1e; ++it1){
        op(*it1);
      }
      return yaRC_ok;
    }
    
    
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2 it2, const iter2 it2e, image1&, image2&)
    {
      // todo : some cases where im1 == im2 && it1 == it2 && it1e == it2e should be optimized
      for(; it1 != it1e && it2 != it2e; ++it1, ++it2){
        *it2 = op(*it1);
      }
      return yaRC_ok;
    }
  };

  //! Specializing of s_apply_op_range for iterators sharing the same offset configuration
  template <>
  struct s_apply_op_range<iterators_same_offset_tag, operator_type_unary>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image2& im2)
    {
      // todo : some cases where im1 == im2 && it1 == it2 && it1e == it2e should be optimized
      for(; it1 != it1e; ++it1){
        const offset o = it1.Offset();
        im2.pixel(o) = op(im1.pixel(o));
      }
      return yaRC_ok;
    }
  };

  //! Specializing of s_apply_op_range for iterators sharing the same offset configuration
  template <>
  struct s_apply_op_range<iterators_same_offset_and_same_images_tag, operator_type_unary>
  {
    //! Unary operation with return value
    template <class op_, class iter1, class iter2, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, iter2, const iter2, image1& im1, image1&)
    {
      typedef typename mpl::if_<
        is_const<image1>,
        typename image1::const_reference,
        typename image1::reference>::type ref_t;    
    
      for(; it1 != it1e; ++it1){
        ref_t p = *it1;
        p = op(p);
      }
      return yaRC_ok;
    }

    template <class op_, class iter1, class iter2, class image1, class image2>
    yaRC operator()(op_&, iter1, const iter1, iter2, const iter2, image1&, image2&)
    {
      YAYI_THROW("This function should never be called : the program is ill-formed !");
    }
  };


  //! Specializing of s_apply_op_range for independant iterators and unary operators without return value
  template <>
  struct s_apply_op_range<iterators_independant_tag, operator_type_unary_no_return>
  {
    template <class op_, class iter1, class image1>
    yaRC operator()(op_& op, iter1 it1, const iter1 it1e, image1&)
    {
      for(; it1 != it1e; ++it1){
        op(*it1);
      }
      return yaRC_ok;
    }

  };
  
  
  /*!@brief Unary operator functor over an image
   *
   * @author Raffi Enficiaud
   */
  struct s_apply_unary_operator
  {

  private:
    template <class iterator_extractor_t, class image_in_out, class op_>
    yaRC apply(image_in_out& im, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor_t> extractor_type;
      
      extractor_type extractor;
      
      typedef typename extractor_type::template results<extractor_type(image_in_out&)>::type result_type;
      
      const result_type iterators(extractor(im));
      typename result_type::first_type  it = iterators.first;
      typename result_type::second_type const ite = iterators.second;
      
      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< op_(typename result_type::first_type::reference_type) >::type range_type;
      
      
      
      // Here we can add more precision on the type of 
      s_apply_op_range<iterators_independant_tag, range_type> op_apply;
      
      return op_apply(op, it, ite, im);
    }

    // TODO : separate the extractors for each image
    template <class iterator_extractor_t, class image_in, class image_out, class op_>
    yaRC apply(image_in& im, image_out& out, op_& op) const
    {
      //BOOST_STATIC_ASSERT(is_void<typename op_::result_type>::value);
      typedef s_iterator_extractor<iterator_extractor_t> extractor_type;
      
      extractor_type extractor;
      
      typedef typename extractor_type::template results<extractor_type(image_in&)>::type itrange_in_t;
      typedef typename extractor_type::template results<extractor_type(image_out&)>::type itrange_out_t;
      
      itrange_in_t const iterators_in(extractor(im));
      typename itrange_in_t::first_type  it = iterators_in.first;
      typename itrange_in_t::second_type const ite = iterators_in.second;
      
      itrange_out_t const iterators_out(extractor(out));
      typename itrange_out_t::first_type  ito = iterators_out.first;
      typename itrange_out_t::second_type const itoe = iterators_out.second;
      
      // test the return of op if it exists ??
      typedef typename s_extract_operator_type::template result< op_(typename itrange_in_t::first_type::reference_type) >::type range_type;
      
      
      switch(iterator_strategy(im, out))
      {
        case eis_window_maximal_same_pointer:
        {
          s_apply_op_range<iterators_same_offset_and_same_images_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);      
        }
        
        case eis_pointer_compatibles:
        case eis_window_maximal_same_offset:
        case eis_offset_compatibles:
        {
          s_apply_op_range<iterators_same_offset_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);
        }
        
        default: 
        {
          s_apply_op_range<iterators_independant_tag, range_type> op_apply;
          return op_apply(op, it, ite, ito, itoe, im, out);
        }
    
      }  
      
    }
    



  public:
    template <class image_in_out, class op_>
    yaRC operator()(image_in_out& im, op_& op) const
    {
      switch(iterator_extractor(im))
      {
      case eic_non_windowed:
        {
          return apply<iterator_choice_strategy_non_windowed_tag>(im, op);
        }
      /*case eic_windowed:
        {
          return apply<image_in_out, iterator_choice_strategy_windowed_tag>(im, op);
        }*/
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(im)));
      }
    }

    template <class image_in, class image_out, class op_>
    yaRC operator()(const image_in& im, image_out& out, op_& op) const
    {
      switch(iterator_extractor(im))
      {
      case eic_non_windowed:
        {
          return apply<iterator_choice_strategy_non_windowed_tag>(im, out, op);
        }
      /*case eic_windowed:
        {
          return apply<image_in_out, iterator_choice_strategy_windowed_tag>(im, op);
        }*/
      default:
        YAYI_THROW("Unsupported case " + any_to_string(iterator_extractor(im)));
      }
    }



  };





}



#endif /* YAYI_APPLY_TO_IMAGE_UNARY_HPP__ */

