/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2010
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_LOWLEVEL_MORPHOLOGY_NEIGHBOR_OPERATORS_HPP__
#define YAYI_LOWLEVEL_MORPHOLOGY_NEIGHBOR_OPERATORS_HPP__

/*!@file
 * This file defines some neighbor operators
 */
 
#include <boost/circular_buffer.hpp>
#include <boost/call_traits.hpp>
#include <algorithm>

namespace yayi { namespace llmm {

  namespace neighborhood_operator_traits {
    struct shiftable_operator;
  }


  /*!@brief Min element over a range
   * The functor returns a reference to the min element over the provided range
   * @author Raffi Enficiaud
   */
  template <class in_t>
  struct s_min_element_subset {
  
    typedef neighborhood_operator_traits::shiftable_operator operator_t;
  
    template <class it>
    typename it::const_reference operator()(it beg, const it e) const {
      DEBUG_ASSERT(beg != e, "Empty neighborhood");
      return *std::min_element(beg, e);
    }
    
    #if 0
    // for future use
    template <class it>
    struct s_shifter {
      typedef it iterator_t;
      typedef typename iterator_t::value_type value_type;
      
      //std::vector<value_type> vect_values;
      boost::circular_buffer<value_type> values;

      
      template <class neigh_t>
      s_shifter(const neigh_t& n) : values(n.get_shift()) {
      
      }
      
      // Initialize the internals
      void init() {
        values.clear();
      }
      
      // 
      const in_t& step(it beg, const it e) {
        values.push_back(*std::min_element(beg, e));
        return *std::min_element(values.begin(), values.end());
      }
      
    
    };
    #endif
  
  };

  /*!@brief Max element over a range
   * The functor returns a reference to the max element over the provided range
   * @author Raffi Enficiaud
   */
  template <class in_t>
  struct s_max_element_subset {
  
    typedef neighborhood_operator_traits::shiftable_operator operator_t;
  
    template <class it>
    typename it::const_reference operator()(it beg, const it e) const {
      DEBUG_ASSERT(beg != e, "Empty neighborhood");
      return *std::max_element(beg, e);
    }
      
  };

  /*!@brief Max minus min value over a range
   * The functor returns the value of the maximum minus minimum over the provided range
   * @author Raffi Enficiaud
   */
  template <class in_t>
  struct s_max_minus_min_element_subset {
  
    typedef neighborhood_operator_traits::shiftable_operator operator_t;
  
    template <class it>
    typename it::value_type operator()(it beg, const it e) const {
      DEBUG_ASSERT(beg != e, "Empty neighborhood");
      typename it::value_type max_e = *beg, min_e = *beg;
      for(++beg; beg != e; ++beg)
      {
        typename it::value_type val = *beg;
        max_e = std::max(val, max_e);
        min_e = std::min(val, min_e);
      }
      return max_e - min_e;
    }
      
  };

  /*!@brief Max minus min value over a range
   * The functor returns the value of the maximum minus minimum over the provided range
   * @author Raffi Enficiaud
   */
  template <class in_t>
  struct s_max_minus_center_element_subset {
  
    typedef neighborhood_operator_traits::shiftable_operator operator_t;
  
    template <class it>
    typename it::value_type operator()(it beg, const it e, typename boost::call_traits<typename it::value_type>::param_type center_value) const {
      static const s_max_element_subset<in_t> elem = s_max_element_subset<in_t>();
      return elem(beg,e) - center_value;
    }
      
  };


  /*!@brief Max minus min value over a range
   * The functor returns the value of the maximum minus minimum over the provided range
   * @author Raffi Enficiaud
   */
  template <class in_t>
  struct s_center_minus_min_element_subset {
  
    typedef neighborhood_operator_traits::shiftable_operator operator_t;
  
    template <class it>
    typename it::value_type operator()(it beg, const it e, typename boost::call_traits<typename it::value_type>::param_type center_value) const {
      static const s_min_element_subset<in_t> elem = s_min_element_subset<in_t>();
      return center_value - elem(beg,e);
    }
      
  };

}}



#endif /* YAYI_LOWLEVEL_MORPHOLOGY_NEIGHBOR_OPERATORS_HPP__ */
