/*
 *  -:- LICENCE -:- 
 * Copyright Raffi Enficiaud 2007-2010
 * 
 * Distributed under the Boost Software License, Version 1.0.
 * (See accompanying file ../../../LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 * 
 *  -:- LICENCE -:- 
 */

#ifndef YAYI_PIXEL_IMAGE_ARITHMETIC_T_HPP__
#define YAYI_PIXEL_IMAGE_ARITHMETIC_T_HPP__

/*!@file
 * This file defines the arithmetic template operations on images
 */
 
#include <boost/call_traits.hpp>
#include <functional>
#include <Yayi/core/yayiPixelProcessing/image_arithmetics.hpp>
#include <Yayi/core/yayiImageCore/include/ApplyToImage_unary_t.hpp>
#include <Yayi/core/yayiImageCore/include/ApplyToImage_binary_t.hpp>
#include <Yayi/core/yayiCommon/common_errors.hpp> // to remove
namespace yayi
{

  template <class in1_t, class in2_t, class out_t>
  struct s_add : 
    std::binary_function<
      typename boost::call_traits<in1_t>::param_type, 
      typename boost::call_traits<in2_t>::param_type, 
      out_t>
  {
    
    out_t operator()(
      typename boost::call_traits<in1_t>::param_type v1, 
      typename boost::call_traits<in2_t>::param_type v2) const throw()
    {
      return static_cast<out_t>(v1 + v2);
    }
  };

  template <class image_in1_, class image_in2_, class image_out_>
  yaRC add_images_t(const image_in1_& imin1, const image_in2_& imin2, image_out_& imo)
  {
    typedef s_add<
      typename image_in1_::pixel_type, 
      typename image_in2_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_binary_operator op_processor;
    
    operator_type op;
    return op_processor(imin1, imin2, imo, op);
  }



  template <class in1_t, class in2_t, class out_t>
  struct s_sub : 
    std::binary_function<
      typename boost::call_traits<in1_t>::param_type, 
      typename boost::call_traits<in2_t>::param_type, 
      out_t>
  {
    
    out_t operator()(
      typename boost::call_traits<in1_t>::param_type v1, 
      typename boost::call_traits<in2_t>::param_type v2) const throw()
    {
      return static_cast<out_t>(v1 - v2);
    }
  
  };

  template <class image_in1_, class image_in2_, class image_out_>
  yaRC subtract_images_t(const image_in1_& imin1, const image_in2_& imin2, image_out_& imo)
  {
    typedef s_sub<
      typename image_in1_::pixel_type, 
      typename image_in2_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_binary_operator op_processor;
    
    operator_type op;
    return op_processor(imin1, imin2, imo, op);
  }



  template <class in1_t, class in2_t, class out_t>
  struct s_mult : 
    std::binary_function<
      typename boost::call_traits<in1_t>::param_type, 
      typename boost::call_traits<in2_t>::param_type, 
      out_t>
  {
    
    out_t operator()(
      typename boost::call_traits<in1_t>::param_type v1, 
      typename boost::call_traits<in2_t>::param_type v2) const throw()
    {
      return static_cast<out_t>(v1 * v2);
    }
  
  };

  template <class image_in1_, class image_in2_, class image_out_>
  yaRC multiply_images_t(const image_in1_& imin1, const image_in2_& imin2, image_out_& imo)
  {
    typedef s_mult<
      typename image_in1_::pixel_type, 
      typename image_in2_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_binary_operator op_processor;
    
    operator_type op;
    return op_processor(imin1, imin2, imo, op);
  }



  template <class in1_t, class in2_t, class out_t>
  struct s_div : 
    std::binary_function<
      typename boost::call_traits<in1_t>::param_type, 
      typename boost::call_traits<in2_t>::param_type, 
      out_t>
  {
    
    out_t operator()(
      typename boost::call_traits<in1_t>::param_type v1, 
      typename boost::call_traits<in2_t>::param_type v2) const throw()
    {
      return static_cast<out_t>(v1 / v2);
    }
  
  };

  template <class image_in1_, class image_in2_, class image_out_>
  yaRC divide_images_t(const image_in1_& imin1, const image_in2_& imin2, image_out_& imo)
  {
    typedef s_div<
      typename image_in1_::pixel_type, 
      typename image_in2_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_binary_operator op_processor;
    
    operator_type op;
    return op_processor(imin1, imin2, imo, op);
  }




  template <class in1_t, class val_t, class out_t>
  struct s_add_constant : 
    std::unary_function<
      typename boost::call_traits<in1_t>::param_type, 
      out_t>
  {  
    typename boost::add_const<val_t>::type value; 
    s_add_constant(typename boost::call_traits<val_t>::param_type p) : value(p) {}
    out_t operator()(typename boost::call_traits<in1_t>::param_type v1) const throw()
    {
      return static_cast<out_t>(v1 + value);
    }
  };
  
  
  template <class image_in1_, class image_out_>
  yaRC add_images_constant_t(const image_in1_& imin1, typename image_in1_::const_reference val, image_out_& imo)
  {
    typedef s_add_constant<
      typename image_in1_::pixel_type, 
      typename image_in1_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_unary_operator op_processor;
    
    operator_type op(val);
    return op_processor(imin1, imo, op);
  }
  

  template <class in1_t, class val_t, class out_t>
  struct s_sub_constant : 
    std::unary_function<
      typename boost::call_traits<in1_t>::param_type, 
      out_t>
  {  
    typename boost::add_const<val_t>::type value; 
    s_sub_constant(typename boost::call_traits<val_t>::param_type p) : value(p) {}
    out_t operator()(typename boost::call_traits<in1_t>::param_type v1) const throw()
    {
      return static_cast<out_t>(v1 - value);
    }
  };
  
  template <class image_in1_, class image_out_>
  yaRC subtract_images_constant_t(
    const image_in1_& imin1, 
    typename boost::call_traits<typename image_in1_::pixel_value>::param_type val, 
    image_out_& imo)
  {
    typedef s_sub_constant<
      typename image_in1_::pixel_type, 
      typename image_in1_::pixel_type, 
      typename image_out_::pixel_type>  operator_type;
    
    s_apply_unary_operator op_processor;
    
    operator_type op(val);
    return op_processor(imin1, imo, op);
  }
  
  


  
  template <class in1_t, class val_t, class out_t>
  struct s_mult_constant : 
    std::unary_function<
      typename boost::call_traits<in1_t>::param_type, 
      out_t>
  {  
    typename boost::add_const<val_t>::type value; 
    s_mult_constant(typename boost::call_traits<val_t>::param_type p) : value(p) {}
    out_t operator()(typename boost::call_traits<in1_t>::param_type v1) const throw()
    {
      return static_cast<out_t>(v1 * value);
    }
  };
  
  
  //! Image multiplication with a constant value
  //! Care should be taken for the multiplication operand type
  template <class image_in1_, class val_t, class image_out_>
  yaRC multiply_images_constant_t(
    const image_in1_& imin1, 
    typename boost::call_traits<val_t>::param_type val, 
    image_out_& imo)
  {
    typedef s_mult_constant<
      typename image_in1_::pixel_type, 
      val_t, 
      typename image_out_::pixel_type>  operator_type;
      
    std::cout << "value = " << (double)val << " " << errors::demangle(typeid(typename boost::call_traits<val_t>::param_type).name()) << std::endl;
    
    s_apply_unary_operator op_processor;
    
    operator_type op(val);
    return op_processor(imin1, imo, op);
  }
  

}

#endif
